//
//  BasicS3.swift
//  BasicS3
//

import Foundation
import CommonCrypto

class BasicS3 {
    fileprivate var awsID:String!
    fileprivate var awsKey:String!
    
    fileprivate let S3Host = "s3.amazonaws.com"
    fileprivate var region:String
    fileprivate var bucket:String
    fileprivate var insecureHost:String { return bucket + "." + S3Host }
    fileprivate var secureHost:String { return S3Host }
    
    fileprivate let EmptyBodySHA256 = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
    
    fileprivate lazy var loginProtectionSpace:URLProtectionSpace = URLProtectionSpace(host: self.secureHost, port: 80, protocol: "HTTP", realm: nil, authenticationMethod: NSURLAuthenticationMethodDefault)

    init (region r: String, bucket b: String) {
        region = r
        bucket = b
    }

    func loadAwsCredentials () -> Bool {
        let credentials = URLCredentialStorage.shared.credentials(for: loginProtectionSpace)
        if credentials!.isEmpty == true { return false }
        
        awsID = credentials!.first!.value.user
        awsKey = credentials!.first!.value.password
        return true
    }
    
    func setAwsCredentials (awsID id:String, awsKey key:String) {
        awsID = id
        awsKey = key
        
        deleteAwsCredentials()

        URLCredentialStorage.shared.set(URLCredential(user: awsID, password: awsKey, persistence: URLCredential.Persistence.permanent), for: loginProtectionSpace)
    }

    func deleteAwsCredentials () {
        URLCredentialStorage.shared.credentials(for: loginProtectionSpace)?.forEach { URLCredentialStorage.shared.remove($1, for: loginProtectionSpace) }
    }

    func credentialsAreLoaded () -> Bool {
        return awsID != nil || awsKey != nil
    }
    
    func copyAwsCredentials (from: BasicS3) {
        awsID = from.awsID
        awsKey = from.awsKey
    }

    fileprivate func signatureKey (_ date:Date) -> Data {
        let kSecret = ("AWS4" + awsKey) as NSString
        let kSecretBytes = kSecret.cString(using: String.Encoding.utf8.rawValue)
        let kSecretLength = kSecret.lengthOfBytes(using: String.Encoding.utf8.rawValue)

        let dateFormatter = DateFormatter()
        dateFormatter.locale = Locale(identifier:"en_US_POSIX")
        dateFormatter.dateFormat = "yyyyMMdd"
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        let dateString = dateFormatter.string(from: date)

        let kDate = hmacSHA256(string: dateString, key: Data(bytes: UnsafeRawPointer(kSecretBytes!), count: kSecretLength))
        let kRegion = hmacSHA256(string: region, key: kDate)
        let kService = hmacSHA256(string: "s3", key: kRegion)
        let kSigning = hmacSHA256(string: "aws4_request", key: kService)
        
        return kSigning
    }
    
    fileprivate func authorization (_ method: String, path: String, headers: inout [String:String], queries: [String:String], bodyHash: String, https: Bool) -> String {
        let date = Date(timeIntervalSinceNow: 0)
        
        let dateFormatter = DateFormatter()
        dateFormatter.locale = Locale(identifier:"en_US_POSIX")
        dateFormatter.dateFormat = "yyyyMMdd'T'HHmmss'Z'"
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        let dateTimeString = dateFormatter.string(from: date)
        
        headers["x-amz-date"] = dateTimeString
        headers["Host"] = https ? secureHost : insecureHost
        
        let canonicalPath = "/" + UriEncode(path, encodeSlash: false)
        let canonicalHeaderKeys = headers.keys.map { $0.lowercased() }.sorted().joined(separator: ";")
        let canonicalHeaders = headers.keys.sorted{ $0.caseInsensitiveCompare($1) == ComparisonResult.orderedAscending }.map { $0.lowercased() + ":" + headers[$0]! }.joined(separator: "\n") + "\n"
        let canonicalQueries = queries.keys.sorted().map { UriEncode($0, encodeSlash: true) + "="  + UriEncode(queries[$0]!, encodeSlash: true) }.joined(separator: "&")
        let canonicalRequest = method + "\n" + canonicalPath + "\n" + canonicalQueries + "\n" + canonicalHeaders + "\n" + canonicalHeaderKeys + "\n" + bodyHash

        dateFormatter.dateFormat = "yyyyMMdd"
        let dateString = dateFormatter.string(from: date)
        let scope = dateString + "/" + region + "/" + "s3" + "/" + "aws4_request"

        let stringToSign = "AWS4" + "-" + "HMAC-SHA256" + "\n" + dateTimeString + "\n" + scope + "\n" + hexString(data: hashSHA256(string: canonicalRequest), uppercase: false)
        let signingKey = signatureKey(date)
        let signature = hexString(data: hmacSHA256(string: stringToSign, key: signingKey), uppercase: false)
        
        let retA = "AWS4" + "-" + "HMAC-SHA256"
        let retB = "Credential=" + awsID + "/" + scope
        let retC = "SignedHeaders=" + canonicalHeaderKeys
        let retD = "Signature=" + signature
        return retA + " " + retB + "," + retC + "," + retD
    }
    
    fileprivate func hashMD5(data: Data) -> Data {
        let dataBytes = (data as NSData).bytes.bindMemory(to: CUnsignedChar.self, capacity: data.count)
        let digestLength = Int(CC_MD5_DIGEST_LENGTH)
        let digest = UnsafeMutablePointer<CUnsignedChar>.allocate(capacity: digestLength)
        defer { digest.deallocate() }

        CC_MD5(dataBytes, CC_LONG(data.count), digest)

        return Data(bytes: digest, count: digestLength)
    }

    fileprivate func hashSHA256(data: Data) -> Data {
        let dataBytes = (data as NSData).bytes.bindMemory(to: CUnsignedChar.self, capacity: data.count)
        let digestLength = Int(CC_SHA256_DIGEST_LENGTH)
        let digest = UnsafeMutablePointer<CUnsignedChar>.allocate(capacity: digestLength)
        defer { digest.deallocate() }
        
        CC_SHA256(dataBytes, CC_LONG(data.count), digest);
        
        return Data(bytes: digest, count: digestLength)
    }

    fileprivate func hashSHA256(string: String) -> Data {
        let data = string.cString(using: String.Encoding.utf8)
        let dataLen = UInt32(string.lengthOfBytes(using: String.Encoding.utf8))
        let digestLen = Int(CC_SHA256_DIGEST_LENGTH)
        let digest = UnsafeMutablePointer<CUnsignedChar>.allocate(capacity: digestLen)
        defer { digest.deallocate() }

        CC_SHA256(data, dataLen, digest);

        return Data(bytes: digest, count: digestLen)
    }

    fileprivate func hmacSHA256(string: String, key: Data) -> Data {
        let keyBytes = (key as NSData).bytes.bindMemory(to: CUnsignedChar.self, capacity: key.count)
        let data = string.cString(using: String.Encoding.utf8)
        let dataLen = Int(string.lengthOfBytes(using: String.Encoding.utf8))
        let digestLen = Int(CC_SHA256_DIGEST_LENGTH)
        let digest = UnsafeMutablePointer<CUnsignedChar>.allocate(capacity: digestLen)
        defer { digest.deallocate() }
        
        CCHmac(CCHmacAlgorithm(kCCHmacAlgSHA256), keyBytes, key.count, data, dataLen, digest);

        return Data(bytes: digest, count: digestLen)
    }

    fileprivate func hexString (data:Data, uppercase:Bool) -> String {
        if data.count == 0 { return "" }

        let hexChars = Array((uppercase ? "0123456789ABCDEF" : "0123456789abcdef").utf8) as [UInt8]
        let buf = UnsafeBufferPointer<UInt8>(start: (data as NSData).bytes.bindMemory(to: UInt8.self, capacity: data.count), count: data.count)
        var output = [UInt8](repeating: 0, count: data.count*2 + 1)
        var ix:Int = 0
        
        for b in buf {
            let hi  = Int((b & 0xf0) >> 4)
            let low = Int(b & 0x0f)
            output[ix] = hexChars[ hi]
            ix += 1
            output[ix] = hexChars[low]
            ix += 1
        }
        
        return String(cString: UnsafePointer(output));
    }
    
    fileprivate func hexString (signedByte:Int8, uppercase:Bool) -> String {
        let ch = UInt8(bitPattern: signedByte)
        let hexChars = (uppercase ? "0123456789ABCDEF" : "0123456789abcdef").cString(using: String.Encoding.utf8)!
        var output = [Int8](repeating: 0, count: 2 + 1)
        
        let hi  = Int((ch & 0xf0) >> 4)
        let low = Int(ch & 0x0f)
        output[0] = hexChars[ hi]
        output[1] = hexChars[low]
        
        return String(cString: UnsafePointer(output))
    }
    
    fileprivate func UriEncode(_ path:String, encodeSlash:Bool) -> String {
        var result:String = ""
        var str = path.cString(using: String.Encoding.utf8)!
        /* Remove \0 from end of cstring */
        str.removeLast()
        
        for ch in str {
            if (48 <= ch && ch <= 57) || (65 <= ch && ch <= 90) || (97 <= ch && ch <= 122) || ch == 95 || ch == 45 || ch == 126 || ch == 46 || (ch == 47 && !encodeSlash) {
                result.append(Character(UnicodeScalar(Int(ch))!))
            }
            else {
                result.append("%" + hexString(signedByte: ch, uppercase: true))
            }
        }
        
        return result
    }
    
    fileprivate func s3Request (method:String, objectPath:String, queries:[String:String], headers argHeaders:[String:String], dataToSend:Data?, secure:Bool, dataReceived:inout Data?, response:inout HTTPURLResponse?, error:inout Error?) -> Bool {
        let dataShaHash = dataToSend != nil ? hexString(data:hashSHA256(data: dataToSend!), uppercase:false) : EmptyBodySHA256
        
        let canonicalQueries = queries.keys.sorted().map { UriEncode($0, encodeSlash: true) + "="  + UriEncode(queries[$0]!, encodeSlash: true) }.joined(separator: "&")
        let pathString = canonicalQueries.isEmpty ? objectPath : objectPath + "?" + canonicalQueries

        var headers = argHeaders
        headers["x-amz-content-sha256"] = dataShaHash
        headers["Authorization"] = authorization(method, path: secure ? bucket + "/" + objectPath : objectPath, headers: &headers, queries: queries, bodyHash: dataShaHash, https: secure)
    
        let session = URLSession(configuration: URLSessionConfiguration.default, delegate: nil, delegateQueue: nil)

        let request = secure ? NSMutableURLRequest(url: URL(string: "https://" + secureHost + "/" + UriEncode(bucket + "/" + objectPath, encodeSlash: false) + (canonicalQueries.isEmpty ? "" : "?" + canonicalQueries))!) : NSMutableURLRequest(url: URL(string: "http://" + insecureHost + "/" + pathString)!)

        request.httpMethod = method
        request.cachePolicy = NSURLRequest.CachePolicy.reloadIgnoringLocalCacheData
        for h in headers.keys {
            request.setValue(headers[h], forHTTPHeaderField: h)
        }
        request.httpBody = dataToSend

        var tmpData:Data?
        var tmpResponse:HTTPURLResponse?
        var tmpError:Error?

        let semaphore = DispatchSemaphore(value: 0)
        session.dataTask(with: request as URLRequest) { d, r, e in
            tmpData = d
            tmpError = e
            tmpResponse = (e == nil) ? (r as! HTTPURLResponse) : nil
            semaphore.signal()
        }.resume()
        semaphore.wait()
        
        dataReceived = tmpData
        response = tmpResponse
        error = tmpError

        return error == nil
    }
    
    func parseS3ObjectResponse (response: HTTPURLResponse, data: Data, objectPath: String) -> BasicS3Object? {
        if response.allHeaderFields["ETag"] == nil && String(data: data, encoding: String.Encoding.utf8)!.contains("<Error><Code>NoSuchKey</Code>") { return nil }
    
        let ret = BasicS3Object()
        ret.data = data
        ret.etag = (response.allHeaderFields["Etag"] as? String)?.replacingOccurrences(of: "\"", with: "")
        let df = DateFormatter()
        df.dateFormat = "EEE, d MMM yyyy HH:mm:ss zzz"
        let lm = response.allHeaderFields["Last-Modified"] as! String?
        ret.lastModified =  lm != nil ? df.date(from: lm!) : nil
        ret.name = objectPath.components(separatedBy: "/").last!
        
        return ret
    }
    
    func getObject (objectPath:String) -> BasicS3Object? {
        var data: Data?
        var response: HTTPURLResponse?
        var error: Error?

        if !s3Request(method: "GET", objectPath: objectPath, queries: [:], headers: [:], dataToSend: nil, secure: true, dataReceived: &data, response: &response, error: &error) {
            return nil
        }

        return parseS3ObjectResponse(response: response!, data: data!, objectPath: objectPath)
    }

    func headObject (objectPath:String) -> BasicS3Object? {
        var data: Data?
        var response: HTTPURLResponse?
        var error: Error?

        if !s3Request(method: "HEAD", objectPath: objectPath, queries: [:], headers: [:], dataToSend: nil, secure: true, dataReceived: &data, response: &response, error: &error) {
            return nil
        }
        
        return parseS3ObjectResponse(response: response!, data: data!, objectPath: objectPath)
    }
    
    func putObject (objectPath:String, data dataToSend:Data, contentType:String, headers argHeaders:[String:String] = [:]) -> Bool {
        let dataMD5Hash = hashMD5(data: dataToSend)

        var headers = argHeaders
        headers["Content-MD5"] = dataMD5Hash.base64EncodedString()
        headers["Content-Type"] = contentType

        var dataReceived: Data?
        var response: HTTPURLResponse?
        var error: Error?

        if !s3Request(method: "PUT", objectPath: objectPath, queries: [:], headers: headers, dataToSend: dataToSend, secure: true, dataReceived: &dataReceived, response: &response, error: &error) {
            return false
        }
        
        return (response?.allHeaderFields["Etag"] as! String?)?.replacingOccurrences(of: "\"", with: "") == hexString(data: dataMD5Hash, uppercase: false)
    }
    
    func deleteObject (objectPath:String) -> Bool {
        var dataReceived: Data?
        var response: HTTPURLResponse?
        var error: Error?

        if !s3Request(method: "DELETE", objectPath: objectPath, queries: [:], headers: [:], dataToSend: nil, secure: true, dataReceived: &dataReceived, response: &response, error: &error) {
            return false
        }
        
        /* s3 does not respond in a way to indicate whether it actually deleted the object */
        return true
    }
    
    func listObjects () -> BasicS3Folder? {
        var token:String? = nil
        let rootFolder = BasicS3Folder()
        var prefixList:[String] = [""]
        
        while !prefixList.isEmpty {
            let pfx = prefixList.popLast()!
            if var folder = rootFolder.navigate(relativePath: pfx) {
                repeat {
                    if !oneIterationToListObjects (prefix: pfx, folder: &folder, continuation: &token) { return nil }
                    prefixList.append(contentsOf: folder.newSubfolders.map({ pfx + $0 + "/" }))
                } while token != nil
            }
            else {
                NSLog("Error navigating to prefix \(pfx)")
            }
        }
        
        return rootFolder
    }

    /* Returns false in case of error */
    fileprivate func oneIterationToListObjects (prefix: String?, folder: inout BasicS3Folder, continuation: inout String?) -> Bool {
        var queries = ["list-type":"2", "delimiter":"/", "prefix":(prefix ?? "")]
        if continuation != nil { queries["continuation-token"] = continuation }
        
        var data: Data?
        var response: HTTPURLResponse?
        var error: Error?

        if !s3Request(method: "GET", objectPath: "", queries: queries, headers: [:], dataToSend: nil, secure: true, dataReceived: &data, response: &response, error: &error) {
            return false
        }
        
        /* Assume that if there was no error, response is not nil either */
        
        continuation = response!.allHeaderFields["NextContinuationToken"] as! String? /* Could be nil */

        return folder.parseXML(xml: data!)
    }
}
